# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


import bpy, os, bmesh, json
from numpy import random as npr
from random import uniform, seed
from mathutils import Vector, Euler
from bpy.types import PropertyGroup, UIList, Panel, Operator, Menu
from bpy.props import *

addon_folder = os.path.dirname(__file__)
slicing_presets_folder = os.path.join(addon_folder, "Slicing_Presets")
icons_dir = os.path.join(addon_folder, "Images")

# Previews Collection for the icons
preview_collections = {}

# get icon ID
def get_custom_icon_id(image_name):
    pcoll = preview_collections["icons"]
    icon = pcoll.get(image_name)
    if icon is not None:
        return icon.icon_id
    print('Error with the image:', image_name)
    return 0


############################################################################################
################################### Slicing Presets  #######################################
############################################################################################

# Slicer Props
def slicer_props():
    props = [
        'name',
        'direction',
        'slicer_enabled',
        'relative_distribution',
        'count', 'angle',
        'offset',
        'position',
        'random_angle',
        'random_offset',
        'random_seed'
    ]
    return props             


# Get Slicing Data
def get_slicing_data(slicing_colls):
    props = slicer_props()
    slicing_data = []
    
    for coll in slicing_colls:
        coll_dict = {}
        for p in props:
            prop = getattr(coll, p)
            if prop is None:
                continue
            if hasattr(prop, '__getitem__'):
                coll_dict[p] = prop[:]
            else:
                coll_dict[p] = prop
        slicing_data.append(coll_dict)
    return slicing_data

# Save Slicing Preset
def save_slicing_preset(slicing_colls, preset_file):
    slicing_data = get_slicing_data(slicing_colls)
    with open(preset_file, 'w') as p:
        json.dump(slicing_data, p, indent=5)
        
# Load a Slicing Preset
def load_slicing_preset(preset_file, coll):
    cutters = []
    with open(preset_file, 'r') as p:
        cutters = json.load(p)
    
    if not cutters:
        return
    
    coll.clear()    
    for c in cutters:
        cutter = coll.add()
        for i in c:
            if hasattr(cutter, i):
                setattr(cutter, i, c[i])

# List of the Slicing Presets
def get_slicing_presets():
    presets = []
    if not os.path.exists(slicing_presets_folder):
        return presets
    for p in os.listdir(slicing_presets_folder):
        if p.lower().endswith(".slicing"):
            presets.append(p.replace(".slicing", ""))
    return presets

# List of the custom slicing Presets
def get_custom_slicing_presets(context):
    presets = []
    custom_folder = custom_presets_dir(context)
    if custom_folder is None:
        return presets
    custom_presets = os.path.join(custom_folder, 'Slicing_Presets')
    if not os.path.exists(custom_presets):
        return presets
    if os.path.samefile(custom_presets, slicing_presets_folder):
        return presets
    presets_list = os.listdir(custom_presets)
    presets = [p[:-8] for p in presets_list if p.lower().endswith(".slicing")]
    return presets

# list of all the presets (custom and builtin)
def get_all_presets(context):
    presets_enum = []
    scn = context.scene
    if not len(scn.animax_slicing_settings):
        return [('_Empty_', '_Empty_', '')]
    settings = scn.animax_slicing_settings[0]
    filter = settings.presets_filter
    a_presets = settings.a_presets
    c_presets = settings.c_presets
    filter = filter.strip().lower()
    presets = get_slicing_presets()
    custom_presets = get_custom_slicing_presets(context)
    a_icon = get_custom_icon_id('a_preset.png')
    c_icon = get_custom_icon_id('c_preset.png')
    
    if a_presets:
        presets = [i for i in presets if filter in i.lower()]
        for i, p in enumerate(presets):
            presets_enum.append(('A_'+p, p, '', a_icon, i))
    if c_presets:
        count = len(presets_enum)
        custom_presets = [i for i in custom_presets if filter in i.lower()]
        for i, p in enumerate(custom_presets):
            presets_enum.append(('C_'+p, p, '', c_icon, i+count))
    presets_enum = presets_enum if presets_enum else [('_Empty_', '_Empty_', '')]
    return presets_enum

# List of the Slicing Presets (Enum)
def slicing_presets_list(self, context):
    presets = get_all_presets(context)
    return presets

# update the active preset displayed when some properties chenges
def update_presets(self, context):
    presets_enum = get_all_presets(context)
    self.slicing_presets = presets_enum[0][0]
    return None

############################################################################################
######################################## Utils  ###########################################
############################################################################################

# check if the object (Target, Source) is valide
def check_ob(context, ob):
    objct = None
    ob_empty = ob is None
    ob_ok = False
    ob_is_mesh = False
    if ob_empty:
        return ob_empty, ob_ok, ob_is_mesh
    if not hasattr(ob, 'name'):
        return ob_empty, ob_ok, ob_is_mesh
    objct = context.scene.objects.get(ob.name)
    if objct is None:
        return ob_empty, ob_ok, ob_is_mesh
    ob_ok = True
    if objct.type == 'MESH':
        ob_is_mesh = True
    return ob_empty, ob_ok, ob_is_mesh

# Duplicate Cutter
def duplicate_cutter(context):
    coll = context.scene.animax_slicing_colls.coll
    index = context.scene.animax_slicing_colls.index
    count = len(coll)
    props = slicer_props()
    cutter = coll[index]
    new_cutter = coll.add()
    for p in props:
        value = getattr(cutter, p)
        setattr(new_cutter, p, value)
    new_cutter.name = cutter.name + ' (Copy)'
    context.scene.animax_slicing_colls.index = count
    
#add slicing props
def add_slicing_props(self, context):
    scn = context.scene
    settings = scn.animax_slicing_settings
    if not len(settings):
        settings.add()
        settings.update()
    return None

# Get the list of selected objects and the active object    
def get_selected_obj(context):
    obs = context.scene.objects
    selected = [ob for ob in obs if ob.select_get()]
    active_ob = context.object
    return [active_ob, selected]

# Restore the stat of selected objects and active object    
def set_selected_obj(context, active_ob, selected):
    context.view_layer.objects.active = active_ob
    obs = context.scene.objects
    for ob in obs:
        ob.select_set(False)
    for ob in selected:
        ob.select_set(True)
    
# Get a new unique name for a property
def get_prop_name(name, names):
    newname = name    
    i = 2
    while newname in names:
        newname = name + str(i)
        i += 1
    return newname

def preview_shards(context, ob, preview):
    if not ob:
        return
    
    if not hasattr(ob, 'name'):
        return
    
    scn = context.scene
    
    if not scn.objects.get(ob.name):
        return
    
    act, sel = get_selected_obj(context)
    pivot = scn.tool_settings.transform_pivot_point    
    try:
        set_selected_obj(context, ob, [ob])
        ob.hide_viewport = False
        ob.hide_set(False)
        bpy.ops.object.mode_set(mode='EDIT')
        scn.tool_settings.transform_pivot_point = 'INDIVIDUAL_ORIGINS'
        bpy.ops.mesh.select_all(action='SELECT')
        if preview:
            bpy.ops.transform.resize(value=(0.8, 0.8, 0.8))
        else:
            bpy.ops.transform.resize(value=(1.25, 1.25, 1.25))
    except Exception as e:
        print('Error in preview shards: ' + str(e))
    finally:    
        scn.tool_settings.transform_pivot_point = pivot
        bpy.ops.object.mode_set(mode='OBJECT')
        set_selected_obj(context, act, sel)
    
def update_prev_shard(self, context):
    preview_shards(context, self.target_object, self.preview)
    return None

# Slicing function
def slice_ob(context, target, planes_co, planes_no, mat_index, remove_doubles):
    scn = context.scene
    set_selected_obj(context, target, [target])
    bpy.ops.object.mode_set(mode='EDIT')
    if remove_doubles:
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.remove_doubles()
    if mat_index:
        target.active_material_index = mat_index
    for i, p in enumerate(planes_co):
        bm = bmesh.from_edit_mesh(target.data)
        ret = bmesh.ops.bisect_plane(bm, geom=bm.verts[:]+bm.edges[:]+bm.faces[:], plane_co=p, plane_no=planes_no[i])
        bmesh.ops.split_edges(bm, edges=[e for e in ret['geom_cut'] if isinstance(e, bmesh.types.BMEdge)])        
        bmesh.update_edit_mesh(target.data)
        bpy.ops.mesh.select_all(action='DESELECT')
        bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
        bpy.ops.mesh.select_non_manifold()
        bpy.ops.mesh.edge_face_add()
        bpy.ops.mesh.faces_shade_flat()
        if mat_index:
            bpy.ops.object.material_slot_assign()
    bpy.ops.object.mode_set(mode='OBJECT')

# Generate the Points (positions) and Normals (Orientations) for the cutting plane(s)
def generate_points(slicing_coll, dimensions, global_offset):
    direction = int(slicing_coll.direction)
    offset = slicing_coll.offset * global_offset
    random_offset = slicing_coll.random_offset
    random_rot = slicing_coll.random_angle
    position = slicing_coll.position
    count = slicing_coll.count
    relative = slicing_coll.relative_distribution
    dim = dimensions[direction] if relative else max(dimensions)
    slice_length = dim / (count+1)
    p = [[0.0, 0.0, 0.0] for _ in range(count)]
    n = [list(slicing_coll.angle) for _ in range(count)]
    
    if slicing_coll.randomize_seed:
        slicing_coll.random_seed = npr.choice(range(100000))
    se = slicing_coll.random_seed    
    seed(se)
            
    for i, k in enumerate(n):
        p[i][direction] = (((slice_length*(i+1)) - (dim/2)) * offset) + position
        if random_offset:
            p[i][direction] *= (1-uniform(0.0, random_offset))
            
        for j in range(3):
            if random_rot[j]:
                k[j] *= ((1-random_rot[j]) +  (random_rot[j] * uniform(-1, 1)))
    
    n = [Euler(k).to_matrix()[0] for k in n]
    
    return p, n
    
############################### Finding The Number Of Loose Parts ###############################
# Thanks to the original author Lemon
#https://blender.stackexchange.com/questions/75332/how-to-find-the-number-of-loose-parts-with-blenders-python-api
    
def MakeVertPaths( verts, edges ):
    #Initialize the path with all vertices indexes
    result = {v.index: set() for v in verts}
    #Add the possible paths via edges
    for e in edges:
        result[e.vertices[0]].add(e.vertices[1])
        result[e.vertices[1]].add(e.vertices[0])
    return result

def FollowEdges( startingIndex, paths ):
    current = [startingIndex]

    follow = True
    while follow:
        #Get indexes that are still in the paths
        eligible = set( [ind for ind in current if ind in paths] )
        if len( eligible ) == 0:
            follow = False #Stops if no more
        else:
            #Get the corresponding links
            next = [paths[i] for i in eligible]
            #Remove the previous from the paths
            for key in eligible: paths.pop( key )
            #Get the new links as new inputs
            current = set( [ind for sub in next for ind in sub] )

def count_islands(obj):
    #Prepare the paths/links from each vertex to others
    paths = MakeVertPaths( obj.data.vertices, obj.data.edges )
    found = True
    n = 0
    while found:
        try:
            #Get one input as long there is one
            startingIndex = next( iter( paths.keys() ) )
            n = n + 1
            #Deplete the paths dictionary following this starting index
            FollowEdges( startingIndex, paths )               
        except:
            found = False
    return n

############################### Add-on Preferences ###############################
# Add-on Preferences
def prefs(context):
    addon = context.preferences.addons.get('ANIMAX')
    if addon is not None:
        return addon.preferences
    return None

# Custom presets folder
def custom_presets_dir(context):
    preferences = prefs(context)
    if preferences is None:
        return None
    if not os.path.exists(preferences.presets_path):
        return None
    return preferences.presets_path

############################################################################################
###################################### Operators ###########################################
############################################################################################

# Add Slicing Planes
class ANIMAX_OT_add_slicer(Operator):
    bl_idname = "animax.add_slicer"
    bl_label = "Add"
    bl_options = {'UNDO'}
    bl_description = "Add a group of slicing planes"
    
    def execute(self, context):
        scn = context.scene
        item = scn.animax_slicing_colls.coll.add()
        item.name = "Slicer"
        scn.animax_slicing_colls.index = len(scn.animax_slicing_colls.coll) -1
        return {'FINISHED'}
    
# Delete selected collection
class ANIMAX_OT_delete_slicer(Operator):
    bl_idname = "animax.delete_slicer"
    bl_label = "Delete"
    bl_description = "Delete the selected slicer"
    
    @classmethod
    def poll(cls, context):
        coll = context.scene.animax_slicing_colls.coll
        return len(coll) > 0
    
    def execute(self, context):
        scn = context.scene
        coll = scn.animax_slicing_colls.coll
        index = scn.animax_slicing_colls.index
        coll.remove(index)        
        if index >0:
            scn.animax_slicing_colls.index = (index - 1)
        return {'FINISHED'}    
    
# Calculate and display the number of shards
class ANIMAX_OT_count_shards(Operator):
    bl_idname = "animax.count_islands"
    bl_label = "Number of shards?"
    bl_options = {'UNDO'}
    bl_description = "Calculate then display the number of shards (loose parts)"
    
    count = 0
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not len(scn.animax_slicing_settings):
            return False
        slicing = scn.animax_slicing_settings[0]
        target = slicing.target_object
        if not hasattr(target, 'name'):
            return False
        if not scn.objects.get(target.name):
            return False
        return True
    
    def invoke(self, context, event):
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        target = slicing.target_object
        self.count = count_islands(target)
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):        
        layout = self.layout
        box = layout.box()        
        box.label(text = 'The number of shards is: ' + str(self.count), icon = 'INFO')
    def execute(self, context):
        return {'FINISHED'}
    
# Slicing
class ANIMAX_OT_slice(Operator):
    bl_idname = "animax.slice"
    bl_label = "Execute Slicing"
    bl_options = {'UNDO'}
    bl_description = "Execute the slicing process"
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not len(scn.animax_slicing_settings):
            return False
        slicing = scn.animax_slicing_settings[0]
        src = slicing.source_object
        if not hasattr(src, 'name'):
            return False
        if not scn.objects.get(src.name):
            return False
        if not len(scn.animax_slicing_colls.coll):
            return False
        if slicing.source_object == slicing.target_object:
            return False
        return True
        
    def execute(self, context):
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        source = slicing.source_object
        target = slicing.target_object
        
        if not hasattr(source, 'name'):
            self.report({'WARNING'}, "Set an Object")
            return {'CANCELLED'}
        
        if not scn.objects.get(source.name):
            self.report({'WARNING'}, 'Object: "'+ source.name+'" Not Found.')
            return {'CANCELLED'}
            
        if not source.type == 'MESH':
            self.report({'WARNING'}, "Only Mesh Objects Supported")
            return {'CANCELLED'}
        
        if not hasattr(target, 'name'):
            target = source.copy()
            target.name = source.name + ' (Sliced)'
            slicing.target_object = target
        else:    
            if not scn.objects.get(target.name):
                target = source.copy()
                target.name = source.name + ' (Sliced)'
                slicing.target_object = target
        
        if not scn.collection.objects.get(target.name):
            scn.collection.objects.link(target)
        # Make sure that the target is visible    
        target.hide_viewport = False
        target.hide_set(False)
        
        # Hide the source object       
        source.hide_render = True
        source.hide_set(True)
        
        old_mesh = None
        if target.data != source.data:
            old_mesh = target.data
        
        # Assign a copy of the source mesh to the target
        target.data = source.data.copy()
        
        # Delete the old mesh data to avoid orphan meshes
        if old_mesh and old_mesh.users == 0:
            bpy.data.meshes.remove(old_mesh)
            
        mat_index = slicing.material_index
        remove_doubles = slicing.remove_doubles
        global_offset = slicing.global_offset
        slicing_colls = scn.animax_slicing_colls.coll
        points, normals = [], []
        for i in slicing_colls:
            if not i.slicer_enabled:
                continue
            p, n = generate_points(i, source.dimensions, global_offset)
            points.extend(p)
            normals.extend(n)
        if points and normals:
            slice_ob(context, target, points, normals, mat_index, remove_doubles)
        
        # a hacky way to refresh the preview
        if slicing.preview:
            slicing.preview = True        

        return {'FINISHED'}
    
# Apply Slicing
class ANIMAX_OT_apply_slices(Operator):
    bl_idname = "animax.apply_slice"
    bl_label = "Apply Slicing"
    bl_options = {'UNDO'}
    bl_description = "Convert the shards to objects"
    
    col_name : StringProperty(description = 'Name of the new collection')
    
    @classmethod
    def poll(self, context):
        scn = context.scene
        if not len(scn.animax_slicing_settings):
            return False
        slicing = scn.animax_slicing_settings[0]
        target = slicing.target_object
        if not hasattr(target, 'name'):
            return False
        if not scn.objects.get(target.name):
            return False
        return True
    
    def invoke(self, context, event):
        scn = context.scene
        collections = bpy.data.collections.keys()
        slicing = scn.animax_slicing_settings[0]
        target = slicing.target_object
        self.col_name = get_prop_name(target.name, collections)
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'New Collection')
        layout.prop(self, 'col_name', text = '')
        layout.separator()
        
    def execute(self, context):
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        target = slicing.target_object
            
        if not target.type == 'MESH':
            self.report({'WARNING'}, "Only Mesh Objects Supported")
            return {'CANCELLED'}
        
        if slicing.preview:
            slicing.preview = False
            #preview_shards(context, target, False)
            
        slicing.target_object = None    
            
        act, sel = get_selected_obj(context)
        
        for coll in target.users_collection:
            try:
                coll.objects.unlink(target)
            except:
                pass
        
        collections = bpy.data.collections.keys()
        collection_name = get_prop_name(self.col_name, collections)
        collection = bpy.data.collections.new(collection_name)
        scn.collection.children.link(collection)
        collection.objects.link(target)
        try:
            set_selected_obj(context, target, [target])
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_all(action='SELECT')
            bpy.ops.mesh.separate(type='LOOSE')
            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
        except Exception as e:
            print('Error in Apply Shards: ' + str(e))
        finally:
            set_selected_obj(context, act, sel)
        return {'FINISHED'}
    
# Duplicate Cutter
class ANIMAX_OT_duplicate_cutter(Operator):
    bl_idname = "animax.duplicate_cutter"
    bl_label = "Duplicate"
    bl_options = {'UNDO'}
    bl_description = "Duplicate The Selected Slicer"
    
    @classmethod
    def poll(cls, context):
        return len(context.scene.animax_slicing_colls.coll)
    
    def execute(self, context):        
        duplicate_cutter(context)
        return {'FINISHED'}
    
# Delete all the Cutters
class ANIMAX_OT_delete_cutters(Operator):
    bl_idname = "animax.delete_cutters"
    bl_label = "Delete All Slicers"
    bl_options = {'UNDO'}
    bl_description = "Delete All The Slicers"
    
    @classmethod
    def poll(cls, context):
        return len(context.scene.animax_slicing_colls.coll)
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        layout.separator()
        layout.label(text = 'Delete all the cutters?', icon = 'QUESTION')
        layout.separator()
    
    def execute(self, context):        
        coll = context.scene.animax_slicing_colls.coll
        coll.clear()
        coll.update()
        return {'FINISHED'}
    
# Enable/Disable all the Cutters
class ANIMAX_OT_enable_cutters(Operator):
    bl_idname = "animax.enable_cutter"
    bl_label = "Enable/Disable"
    bl_options = {'UNDO'}
    bl_description = "Enable/Disable all the Slicers"
    
    action : StringProperty()
    
    @classmethod
    def poll(cls, context):
        return len(context.scene.animax_slicing_colls.coll)
    
    def execute(self, context):
        coll = context.scene.animax_slicing_colls.coll
        index = context.scene.animax_slicing_colls.index
        
        count = len(coll)
        seq = [True]*count if self.action == 'ENABLE' else [False]*count
        
        if self.action == 'SOLO':
            seq[index] = True
        coll.foreach_set('slicer_enabled', seq)
        coll.update()
        return {'FINISHED'}
    
# Add the slicing properties
class ANIMAX_OT_add_slicing_props(Operator):
    bl_idname = "animax.add_slicing_props"
    bl_label = "Add Slicing Properties"
    bl_options = {'UNDO'}
    bl_description = "Add the necessary properties for the slicing functionality"
    
    @classmethod
    def poll(cls, context):
        return not len(context.scene.animax_slicing_settings)
    
    def execute(self, context):
        scn = context.scene
        settings = scn.animax_slicing_settings
        if not len(settings):
            settings.add()
            settings.update()
        return {'FINISHED'}
    
# Move SLICER
class ANIMAX_OT_move_slicer(Operator):
    bl_idname = "animax.move_slicer"
    bl_label = "Move"
    bl_options = {'UNDO'}
    bl_description = "Move the selected slicer"
    
    direction : StringProperty(name = 'Direction', default = 'UP')
    
    @classmethod
    def poll(cls, context):
        return len(context.scene.animax_slicing_colls.coll) > 1
    
    def execute(self, context):
        coll = context.scene.animax_slicing_colls.coll
        index = context.scene.animax_slicing_colls.index
        
        if self.direction == 'UP':
            if index > 0:
                coll.move(index, index-1)
                context.scene.animax_slicing_colls.index -= 1
        elif self.direction == 'DOWN':
            if index < len(coll)-1:
                coll.move(index, index+1)
                context.scene.animax_slicing_colls.index += 1
                
        return {'FINISHED'}
    
    
    
########################################## Presets ########################################
    
# Save a Slicing Preset
class ANIMAX_OT_save_slicing_preset(Operator):
    bl_idname = "animax.save_slicing_preset"
    bl_label = "Save a New Preset"
    bl_options = {'UNDO'}
    bl_description = "Save the current slicing data as a preset"
    
    name : StringProperty(
        name = 'Name',
        default = 'My Preset',
        description = 'Name of the new preset',
    )
    
    @classmethod
    def poll(cls, context):
        return len(context.scene.animax_slicing_colls.coll)
        
    def invoke(self, context, event):
        presets = get_custom_slicing_presets(context)
        self.name = get_prop_name('My_Preset', presets)
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        col = layout.column()
        
        custom_folder = custom_presets_dir(context)
        if custom_folder is None:
            preferences = prefs(context)
            col.label(text = 'You need to select the presets folder first', icon = 'INFO')
            col.label(text = 'Or you can set it up from the preferences', icon = 'BLANK1')
            col.prop(preferences, 'presets_path', text = '')
            return
        
        name = bpy.path.clean_name(self.name)
        presets = [i.lower() for i in get_custom_slicing_presets(context)]
        col.alert = name.lower() in presets or not name.strip()
        col.label(text = 'Name of the new Preset')
        col.prop(self, 'name', text = '')
        if name.lower() in presets:
            col = layout.column()
            msg = 'The name "' + name + '" already exists!'
            col.label(text = msg, icon = 'ERROR')
            col.label(text = 'Press OK if you want to overide the preset.', icon = 'INFO')
        
    def execute(self, context):        
        slicing_colls = context.scene.animax_slicing_colls.coll
        name = bpy.path.clean_name(self.name)
        
        custom_folder = custom_presets_dir(context)
        if custom_folder is None:
            self.report({'WARNING'}, 'Custom presets folder not found')
            return {'CANCELLED'}
        presets_folder = os.path.join(custom_folder, 'Slicing_Presets')
        if not os.path.exists(presets_folder):
            try:
                os.makedirs(presets_folder)
            except Exception as e:
                self.report({'WARNING'}, 'Creating the presets folder failed: ' + str(e))
                return {'CANCELLED'}
        
        if not name.strip():
            self.report({'WARNING'}, 'Wrong name')
            return {'CANCELLED'}
            
        preset = os.path.join(presets_folder, name + ".slicing")
        save_slicing_preset(slicing_colls, preset)
        
        if os.path.exists(preset):
            self.report({'INFO'},name + ".slicing" + ' : Saved Successfully!')
        else:
            self.report({'WARNING'}, 'Something Went Wrong! Preset Not Saved!')
            
        return {'FINISHED'}
    
# Load a Slicing Preset
class ANIMAX_OT_load_slicing_preset(Operator):
    bl_idname = "animax.load_slicing_preset"
    bl_label = "Load"
    bl_options = {'UNDO'}
    bl_description = "Load the the selected Preset"
    
    @classmethod
    def poll(cls, context):
        if not len(context.scene.animax_slicing_settings):
            return False
        slicing = context.scene.animax_slicing_settings[0]
        preset = slicing.slicing_presets
        return preset != '_Empty_'
    
    def execute(self, context):
        slicing = context.scene.animax_slicing_settings[0]
        preset = slicing.slicing_presets + '.slicing'
        
        if preset.startswith('C_'):
            custom_folder = custom_presets_dir(context)
            presets_folder = os.path.join(custom_folder, 'Slicing_Presets')
        elif preset.startswith('A_'):
            presets_folder = slicing_presets_folder
            
        preset = preset[2:]
        preset_file = os.path.join(presets_folder, preset)
        if not os.path.exists(preset_file):
            self.report({'WARNING'}, "File doesn't exist")
            return {'CANCELLED'}
        
        coll = context.scene.animax_slicing_colls.coll
        context.scene.animax_slicing_colls.index = 0
        load_slicing_preset(preset_file, coll)
        return {'FINISHED'}    
    
# Delete a Slicing Preset
class ANIMAX_OT_delete_slicing_preset(Operator):
    bl_idname = "animax.delete_slicing_preset"
    bl_label = "Delete"
    bl_options = {'UNDO'}
    bl_description = "Delete the selected Preset"
    
    @classmethod
    def poll(cls, context):
        if not len(context.scene.animax_slicing_settings):
            return False
        slicing = context.scene.animax_slicing_settings[0]
        preset = slicing.slicing_presets
        return preset != '_Empty_'
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        slicing = context.scene.animax_slicing_settings[0]
        preset = slicing.slicing_presets
        preset = preset[2:]
        layout = self.layout
        col = layout.column()
        col.separator()
        col.label(text = 'Delete  "'+ preset +'"  from your Presets?', icon = 'QUESTION' )        
        col.separator()
    
    def execute(self, context):
        slicing = context.scene.animax_slicing_settings[0]
        preset = slicing.slicing_presets + '.slicing'
        
        if preset.startswith('C_'):
            custom_folder = custom_presets_dir(context)
            presets_folder = os.path.join(custom_folder, 'Slicing_Presets')
        elif preset.startswith('A_'):
            presets_folder = slicing_presets_folder
            
        preset = preset[2:]
        preset_file = os.path.join(presets_folder, preset)
        if not os.path.exists(preset_file):
            self.report({'WARNING'}, "File doesn't exist")
            return {'CANCELLED'}
        
        try:
            os.remove(preset_file)
            presets = get_all_presets(context)
            if presets:
                slicing.slicing_presets = presets[0][0]
        except Exception as e:
            self.report({'WARNING'}, str(e))
            return {'CANCELLED'}
        return {'FINISHED'}
    
    
############################################################################################
############################ Property Group for the UI list#################################
############################################################################################
def is_mesh_poll(self, object):
    return object.type == 'MESH'

class SlicingSettings(PropertyGroup):
    preview : BoolProperty(
        default = True,
        description = 'Preview the shards',
        update = update_prev_shard
    )
    remove_doubles : BoolProperty(
        description = 'Merge the overlapping vertices before fracturing'
    )
    global_offset : FloatProperty(
        name = 'Global Offset',
        default = 1.0,
        min = 0.01,
        max = 10000.0,
        description = 'Offset Multiplier, affect the offset of all the cutters'
    )
    material_index : IntProperty(
        default = 0,
        min = 0,
        max = 1000,
        description = 'Material index for interior faces'
    )
    expand_slicer_panels : BoolVectorProperty(
        size = 4,
        description = 'Expand/Collapse This Panel'
    )
    slicing_presets : EnumProperty(
        name = 'Presets',
        items = slicing_presets_list,
        description = 'Slicing Presets'
    )
    source_object : PointerProperty(
        name = 'Source Object',
        type = bpy.types.Object,
        description = 'Object to Slice',
        poll = is_mesh_poll
    )
    
    target_object : PointerProperty(
        name = 'Target Object',
        type = bpy.types.Object,
        description = 'Object to transfer the slicing data to'
    )
    a_presets : BoolProperty(
        name = 'Built-in Presets',
        default = True,
        update = update_presets,
        description = 'Display the presets that are shipped with the add-on'
    )
    c_presets : BoolProperty(
        name = 'Custom Presets',
        default = True,
        update = update_presets,
        description = 'Display the custom presets'
    )
    presets_filter : StringProperty(
        name = 'Presets Filter',
        update = update_presets,
        description = 'Live search filtering string',
        options = {'TEXTEDIT_UPDATE'}
    )

class SlicingColl(PropertyGroup):
    name : StringProperty(default = "Slicer")
    direction : EnumProperty(
        name = 'Direction',
        items = (
            ('0','X',''),
            ('1','Y',''),
            ('2','Z','')
        ),
        description = 'Direction of the cutters array'
    )
    slicer_enabled : BoolProperty(
        default = True,
        description = 'Enable/Disable this Slicer'
    )
    relative_distribution : BoolProperty(
        default = True,
        description = "Distribute the cutters relative to the object's dimensions"
    )
    randomize_seed : BoolProperty(
        default = False,
        description = "Randomize the Random Seed every time you hit Execute"
    )
    count : IntProperty(
        default = 10,
        min = 1,
        max = 10000,
        description = 'Number of cutters'
    )
    angle : FloatVectorProperty(
        name = 'Angle',
        default = (0.0, 0.0, 0.0),
        subtype = 'EULER',
        description = 'Cutter Angle'
    )
    offset : FloatProperty(
        name = 'Offset',
        default = 1.0,
        min = 0.01,
        max = 10000.0,
        description = 'Offset the cutters position'
    )
    position : FloatProperty(
        name = 'Position',
        default = 0.0,
        min = -10000.0,
        max = 10000.0,
        description = 'Position of the cutters'
    )
    random_angle : FloatVectorProperty(
        default = (0.0,0.0,0.0),
        min = 0.0,
        max = 1.0,
        description = 'Randomize the cutters rotation'
    )
    random_offset : FloatProperty(
        default = 0.0,
        min = 0.0,
        max = 1.0,
        description = 'Randomize the cutters offset'
    )
    random_seed : IntProperty(
        default = 1,
        min = 0,
        max = 1000000,
        description = 'Random seed'
    )
    
class SlicingGroup(PropertyGroup):
    coll : CollectionProperty(type=SlicingColl)
    index : IntProperty()
    
############################################################################################
####################################### The UI #############################################
############################################################################################

# Template for seome panels Having the same attributs
class TemplatePanel:
    bl_parent_id = "ANIMAX_PT_slice"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    
    @classmethod
    def poll(self, context):
        if not context.scene.animax_tabs == 'FRACTURE':
            return False
        if not len(context.scene.animax_slicing_settings):
            return False
        return True
    
# UI list Slicing
class SCENE_UL_animax_slicing(UIList):
   
   def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
       row = layout.row()
       row.prop(item, "name", text="", emboss=False)
       row = layout.row()
       row.alignment = 'RIGHT'
       row.prop(item, "slicer_enabled", text="")
    
# Slicing Settings
class SlicingMenu(Menu):
    bl_idname = "ANIMAX_MT_slicing"
    bl_label = "Settings"
    bl_description = "Slicing Settings & Options"        

    def draw(self, context):
        scn = context.scene
        layout = self.layout
        layout.label(text = "Settings", icon='TOOL_SETTINGS')
        layout.separator()
        layout.separator()
        layout.operator("animax.duplicate_cutter", icon='DUPLICATE')
        layout.separator()
        layout.separator()
        layout.operator("animax.delete_cutters", icon='TRASH')
        layout.separator()
        layout.separator()
        layout.operator("animax.enable_cutter", text = 'Enable All', icon='CHECKBOX_HLT').action = 'ENABLE'
        layout.operator("animax.enable_cutter", text = 'Disable All', icon='CHECKBOX_DEHLT').action = 'DISABLE'
        layout.operator("animax.enable_cutter", text = 'Solo Selected', icon='RESTRICT_SELECT_OFF').action = 'SOLO'
    
# Slicing Panel
class ANIMAX_PT_slice(Panel):
    bl_label = "Mesh Slicing"
    bl_parent_id = "ANIMAX_PT_main"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "ANIMAX"
    
    @classmethod
    def poll(self, context):
        return context.scene.animax_tabs == 'FRACTURE'
        
    def draw(self, context):
        layout = self.layout        
        scn = context.scene
        if not len(scn.animax_slicing_settings):
            layout.operator("animax.add_slicing_props", icon = 'PRESET_NEW')
            return
                    
# Input Mesh Panel
class ANIMAX_PT_input_mesh(TemplatePanel, Panel):
    bl_label = "Input Mesh"
    bl_options = {"DEFAULT_CLOSED"}
        
    def draw(self, context):
        layout = self.layout        
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        src = slicing.source_object
        src_empty, src_ok, src_is_mesh = check_ob(context, src)
                    
        box = layout.box()
        box.alert = True if (not src_ok or not src_is_mesh) and not src_empty else False
        col = box.column()
        col.prop(slicing, 'source_object', text = '')
        if src_empty:
            return
        if not src_ok :
            col.label(text = 'Object Not Found', icon = 'ERROR')
        elif not src_is_mesh:
            col.label(text = 'Only Mesh Objects Supported', icon = 'ERROR')
        if src_is_mesh and src_ok:
            if len(src.modifiers):
                box = col.box()
                box.label(text = 'Apply the modifiers for better results', icon = 'INFO')
            if not src.scale == Vector((1,1,1)):
                box = col.box()
                box.label(text = 'Apply the scale of the Source', icon = 'INFO')    
                box.label(text = 'object for better results', icon = 'BLANK1')    
            if not src.data.use_auto_smooth:
                box = col.box()
                box.label(text = 'For smooth shaded objects enable', icon = 'INFO')
                box.label(text = '"Auto Smooth" for better results', icon = 'BLANK1')
                box.prop(src.data, 'use_auto_smooth', icon ='BLANK1')
                
                
# Mesh Cutting Panel
class ANIMAX_PT_cutting(TemplatePanel, Panel):
    bl_label = "Mesh Cutting"
    bl_options = {"DEFAULT_CLOSED"}
    
    def draw(self, context):
        layout = self.layout        
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        a_icon = get_custom_icon_id('a_preset.png')
        c_icon = get_custom_icon_id('c_preset.png')
        box = layout.box()
        col = box.column(align = True)
        row = col.row(align = True)
        row.prop(slicing, 'a_presets', text = '', icon_value = a_icon)
        row.prop(slicing, 'c_presets', text = '', icon_value = c_icon)
        row.prop(slicing, 'presets_filter', text = '', icon = 'VIEWZOOM')
        
        row = col.row(align = True)
        row.prop(slicing, 'slicing_presets', text = '')
        row.operator("animax.load_slicing_preset", icon='FILE_PARENT', text = '')
        row.operator("animax.save_slicing_preset", icon='ADD', text = '')
        row.operator("animax.delete_slicing_preset", icon='REMOVE', text = '')
        row = box.row()
        row.template_list("SCENE_UL_animax_slicing", "coll", scn.animax_slicing_colls, "coll", scn.animax_slicing_colls, "index", rows = 5)  
        col = row.column(align=True)
        col.operator("animax.add_slicer", icon='ADD', text = '')
        col.operator("animax.delete_slicer", icon='REMOVE', text = '')
        col.separator()
        col.menu("ANIMAX_MT_slicing", text = '', icon = 'DOWNARROW_HLT')
        col.separator()
        col.operator("animax.move_slicer", icon='TRIA_UP', text = '').direction = 'UP'
        col.operator("animax.move_slicer", icon='TRIA_DOWN', text = '').direction = 'DOWN'
        col = box.column()
        coll = scn.animax_slicing_colls.coll
        index = scn.animax_slicing_colls.index
        if len(coll):
            slicer = coll[index]
            box1 = col.box()                    
            box1.prop(slicing, "expand_slicer_panels", index = 3,
            icon="TRIA_DOWN" if slicing.expand_slicer_panels[3] else "TRIA_RIGHT",
            icon_only=True, emboss = False, text = 'Slicer Properties')                    
            if slicing.expand_slicer_panels[3]:
                row = box1.row()
                row.prop(slicer, 'direction', expand = True)
                box1.prop(slicer, 'count', text = 'Count')
                row = box1.row(align=True)
                col = row.column()
                col.prop(slicer, 'angle', text = 'Angle')
                col2 = row.column()
                col2.scale_x = 0.8
                col2.prop(slicer, 'random_angle', text = 'Randomize')
                row = box1.row(align=True)
                col = row.column(align = True)
                col.label(text = 'Offset:')
                col.prop(slicer, 'offset', text = '')
                col2 = row.column(align = True)
                col2.scale_x = 0.8
                col2.label(text = 'Randomize:')
                col2.prop(slicer, 'random_offset', text = '')
                row = box1.row(align=True)
                col = row.column(align = True)
                col.label(text = 'Random Seed:')
                col.prop(slicer, 'random_seed', text = '')
                col2 = row.column(align = True)
                col2.scale_x = 0.8
                col2.label(text = 'Randomize:')
                col2.prop(slicer, 'randomize_seed', text = '', icon = 'FILE_REFRESH')
                box1.prop(slicer, 'position', text = 'Position')
                box1.prop(slicer, 'relative_distribution', text = 'Relative Distribution')
            
            col = box.column()
            col.prop(slicing, 'global_offset', text = 'Global Ofsset')
            row = box.row()
            row.prop(slicing, 'material_index', text = 'Material')
            row.prop(slicing, 'remove_doubles', text = 'Remove Doubles', icon = 'GROUP_VERTEX')
            col = box.column()
            col.scale_y = 1.75
            col.operator("animax.slice", text = 'Execute', icon = 'MOD_DECIM')
            
# Output Mesh Panel
class ANIMAX_PT_output_mesh(TemplatePanel, Panel):
    bl_label = "Output Mesh"
    bl_options = {"DEFAULT_CLOSED"}
    
    def draw(self, context):
        layout = self.layout        
        scn = context.scene
        slicing = scn.animax_slicing_settings[0]
        src = slicing.source_object
        src_empty, src_ok, src_is_mesh = check_ob(context, src)
        target = slicing.target_object
        target_empty, target_ok, target_is_mesh = check_ob(context, target)
        
        box = layout.box()
        box.alert = True if (not target_ok or not target_is_mesh) and not target_empty else False
        if not target_empty:
            if target == src:
                box.alert = True
        
        col = box.column()
        col.prop(slicing, 'target_object', text = '')
        if not target_empty:
            if not target_ok:
                col.label(text = 'Object Not Found', icon = 'ERROR')
            elif target == src:
                col.label(text = 'The Input Mesh Is The Output!', icon = 'ERROR')
            elif not target_is_mesh:
                col.label(text = 'Only Mesh Objects Supported', icon = 'ERROR')
            else:
                row = box.row()
                row.prop(slicing, 'preview', text = 'Preview shards', icon = 'MOD_ARRAY')
                row.operator("animax.count_islands", icon = 'INFO')
                col = box.column()
                col.scale_y = 1.5
                col.operator("animax.apply_slice", text = 'Apply', icon = 'MOD_BOOLEAN')
    
    
############################################################################################
################################ Register/Unregister #######################################
############################################################################################

classes = (
    ANIMAX_OT_add_slicer,
    ANIMAX_OT_delete_slicer,
    ANIMAX_OT_count_shards,
    ANIMAX_OT_slice,
    ANIMAX_OT_apply_slices,
    ANIMAX_OT_duplicate_cutter,
    ANIMAX_OT_delete_cutters,
    ANIMAX_OT_enable_cutters,
    ANIMAX_OT_add_slicing_props,
    ANIMAX_OT_move_slicer,
    ANIMAX_OT_save_slicing_preset,
    ANIMAX_OT_load_slicing_preset,
    ANIMAX_OT_delete_slicing_preset,
    SlicingSettings,
    SlicingColl,
    SlicingGroup,
    SCENE_UL_animax_slicing,
    SlicingMenu,
    ANIMAX_PT_slice,
    ANIMAX_PT_input_mesh,
    ANIMAX_PT_cutting,
    ANIMAX_PT_output_mesh,
)

def register():
    from bpy.utils import register_class, previews
    for cls in classes:
        register_class(cls)
        
    bpy.types.Scene.animax_slicing_settings = CollectionProperty(type = SlicingSettings)
    bpy.types.Scene.animax_slicing_colls = PointerProperty(type = SlicingGroup)
    
    # Previews for the icons
    pcoll = previews.new()
    for i in os.listdir(icons_dir):
        if not i.lower().endswith('.png'):
            continue
        pcoll.load(i, os.path.join(icons_dir, i), 'IMAGE')
    preview_collections["icons"] = pcoll

def unregister():
    from bpy.utils import unregister_class, previews
    for cls in reversed(classes):
        unregister_class(cls)

    del bpy.types.Scene.animax_slicing_settings
    del bpy.types.Scene.animax_slicing_colls
    
    for pcoll in preview_collections.values():
        previews.remove(pcoll)
    preview_collections.clear()

if __name__ == "__main__":
    register()